% $Id: bnlik.pl,v 1.5 2004/06/14 20:16:13 jc Exp $

% for memberchk/2
% for member/2
% for select/3
:- use_module(library(lists)).


% Algorithm:
% Basically variable elimination, but avoids doing factor multiplication
% Also childless uninstantiated variables are removed (since our factors
% are always CPTs.

% need to add some reasonably clever variables ordering!!

/*---------------------------------------------------------------------
  Predicate: likelihood(+BN,+VarsVals,-Prob)
       Args: BN a Bayes net as a list of families
             VarsVals: A list of the form
	     [var1-[val11,val12],var2-[var21],..]
	     listing variables in the BN and their permitted values,
	     ie the values we must sum over
             Prob: Probability of that instantiation
   Succeeds: exactly once
   Comments: Inst is a list each element of form Var-PossValues
             Pull out the instantiated variables first
	     Throw away factors where the nonparent is (completely)
	     uninstantiated
---------------------------------------------------------------------*/
likelihood(BN,VarsVals,Prob) :-
	write(likelihood(BN,VarsVals,Prob)),nl,
	classify_vars(VarsVals,[],[],[],UnInst,CompUnInst,Inst),     % 3 sorts of variables
	rmv_uninst_nonparents(BN,CompUnInst,UnInst,BNRed,UnInstRed), % see above
	split_factors(BNRed,Inst,[],[],IFactors,UFactors),           % split off instantiated factors
	factor_mult(IFactors,Inst,1.0,Prob1),                        % and multiply them
	(
	  Prob1 = 0.0
	->
	  Prob = 0.0
	;
	  write(variable_sum(UnInstRed,Inst,UFactors,Prob2)), nl,    % sum out remaining variables
	  variable_sum(UnInstRed,Inst,UFactors,Prob2),               % sum out remaining variables
	  Prob is Prob1*Prob2
	).

/*---------------------------------------------------------------------
  Predicate: variable_sum(+PartialInst,+Inst,+BN,-Prob)
       Args: PartialInst: A list of the form
             [var1-[val11,val12],var2-[var21],..]
	     listing variables in the BN and their permitted values
	     Inst: An instantiation of variables in BN of the form
	     [var3-val13,var4-val42..]
	     BN a Bayes net as a list of families
             Prob: Sum of the probabilities permitted by
	     PartialInst and Inst
   Succeeds: exactly once
   Comments: Each variable in BN occurs either in PartialInst or Inst
             Before doing summation, find any completely instantiated
	     factors (IFactors) and pull them out of the sum.
---------------------------------------------------------------------*/
variable_sum([],Inst,BN,Prob) :-          % Inst now has a val for each var 
	factor_mult(BN,Inst,1.0,Prob).    % so compute simple product
variable_sum([Var-Values|Rest],Inst,BN,Prob) :-
	var_only_unist(BN,Var,Inst,[],[],VarFacts,Others), % once Var-Val is added to inst VarFacts will be inst
	value_sum(Values,Var,Rest,Inst,VarFacts,Others,0.0,Prob).
	
/*---------------------------------------------------------------------
  Predicate: value_sum(+Vals,+Var,+Rest,+Inst,+ProbIn,-ProbOut)
       Args: Vals: a list of values for the variable Var
             Var: the name of a variable
             Rest: A list of the form
             [var1-[val11,val12],var2-[var21],..]
	     listing variables in the BN and their permitted values
	     Inst: An instantiation of variables in BN of the form
	     [var3-val13,var4-val42..]
	     BN a Bayes net as a list of families
             ProbIn: Accumulator input
	     ProbOut: Result of summing over all the values Vals for
	     variable Var, given constraints on other variables in Rest and
	     Inst (modulo ProbIn)
   Succeeds: exactly once
   Comments: Each variable in BN (except Var) occurs either in
             PartialInst or Inst
	     For each Val form a temporary instantiation, and (at (1))
	     sum over the other variables
---------------------------------------------------------------------*/
value_sum([],_Var,_Rest,_Inst,_VarFacts,_Others,Out,Out).
value_sum([Val|Vals],Var,Rest,Inst,VarFacts,Others,In,Out) :-
	factor_mult(VarFacts,[Var-Val|Inst],1.0,Prob1),
	(
	  Prob1 = 0.0
	->
	  Prob = 0.0
	;
	  variable_sum(Rest,[Var-Val|Inst],Others,Prob2), % (1), see above
	  Prob is Prob1 * Prob2
	),
	Mid is In + Prob,
	value_sum(Vals,Var,Rest,Inst,VarFacts,Others,Mid,Out).

/*---------------------------------------------------------------------
  Predicate: factor_mult(+BN,+Inst,+ProbIn,-ProbOut)
       Args: Inst: An instantiation of variables in BN of the form
	     [var3-val13,var4-val42..]
	     BN a Bayes net as a list of families
             ProbIn: Accumulator input
	     ProbOut: Result of multiply all factors in BN for the full
	     instantiation Inst (modulo ProbIn)
   Succeeds: exactly once
   Comments: 
---------------------------------------------------------------------*/
factor_mult(_BN,_Inst,0.0,0.0) :- !. % short-cut, green cut
factor_mult([],_Inst,Out,Out).
factor_mult([Child-Parents|Rest],Inst,In,Out) :-
	memberchk(Child-Val,Inst),           % value for child
	map_memberchk(Parents,Inst,Vals),    % value for parents
	cpt(Child-Val,Parents-Vals,Prob),    % look up prob
	Mid is In * Prob,                    % multiply
	factor_mult(Rest,Inst,Mid,Out).      % and continue





















%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% UTILITIES FROM NOW ON
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/*---------------------------------------------------------------------
  Predicate: classify_vars(+VarsVals,+InUnInst,+InCompUnInst,+InInst,-OutUnInst,-OutCompUnInst,-OutInst)
       Args: 
   Succeeds: exactly once
   Comments: 
---------------------------------------------------------------------*/
classify_vars([],UnInst,CompUnInst,Inst,UnInst,CompUnInst,Inst).
classify_vars([Var-[Val]|Rest],InUnInst,InCompUnInst,InInst,OutUnInst,OutCompUnInst,OutInst) :-
	!,        % found an instantiated variable, change from [Val] to Val
	classify_vars(Rest,InUnInst,InCompUnInst,[Var-Val|InInst],OutUnInst,OutCompUnInst,OutInst).
classify_vars([Var-Vals|Rest],InUnInst,InCompUnInst,InInst,OutUnInst,OutCompUnInst,OutInst) :-
	all_vals(Var,Vals),
	!,        % found an completely uninstantiated variable
	classify_vars(Rest,[Var-Vals|InUnInst],[Var-Vals|InCompUnInst],InInst,OutUnInst,OutCompUnInst,OutInst).
classify_vars([VarVals|Rest],InUnInst,InCompUnInst,InInst,OutUnInst,OutCompUnInst,OutInst) :-
	classify_vars(Rest,[VarVals|InUnInst],InCompUnInst,InInst,OutUnInst,OutCompUnInst,OutInst).


/*---------------------------------------------------------------------
  Predicate: rmv_uninst_nonparents(+BN,+CompUnInst,+UnInst,-BNRed,-UnInstRed)
       Args: BN a Bayes net as a list of families
   Succeeds: exactly once
   Comments: 
---------------------------------------------------------------------*/
rmv_uninst_nonparents(BN,CompUnInst,UnInst,BNRed,UnInstRed):-
	select(Var-_Vals,CompUnInst,CompUnInstMid),   % if you can find a completely uninstantiated variable
	\+ has_child(Var,BN),                   % with no children
	select(Var-_Parents,BN,BNMid),          % remove its family
	select(Var-_Vals,UnInst,UnInstMid),     % remove from list of variables
	!,
	rmv_uninst_nonparents(BNMid,CompUnInstMid,UnInstMid,BNRed,UnInstRed). % and keep trying
rmv_uninst_nonparents(BN,_CompUnInst,UnInst,BN,UnInst). % can't find any more, so retrn

/*---------------------------------------------------------------------
  Predicate: has_child(+Var,+BN)
       Args: Var a variable name 
             BN a Bayes net as a list of families
   Succeeds: at most once
   Comments: plod through parents looking for Var
---------------------------------------------------------------------*/
has_child(Var,BN) :-
	member(_Child-Parents,BN),
	member(Var,Parents),
	!.                         % prevent pointless backtracking.

/*---------------------------------------------------------------------
  Predicate: split_factors(+BN,+Inst,+IIn,+UIn,-IFactors,-UFactors)
       Args: BN a Bayes net as a list of families
             ...
   Succeeds: exactly once
   Comments: Given a BN and an Inst, split BN into instantiated factors
             and those that are uninstantiated
---------------------------------------------------------------------*/
split_factors([],_Inst,BN1,BN2,BN1,BN2).
split_factors([Child-Parents|Rest],Inst,In1,In2,Out1,Out2) :-
	instantiated([Child|Parents],Inst),
	!,
	split_factors(Rest,Inst,[Child-Parents|In1],In2,Out1,Out2).
split_factors([Family|Rest],Inst,In1,In2,Out1,Out2) :-
	split_factors(Rest,Inst,In1,[Family|In2],Out1,Out2).

/*---------------------------------------------------------------------
  Predicate: instantiated(Variables,Inst).
       Args: 
   Succeeds: at most once
   Comments: 
---------------------------------------------------------------------*/
instantiated([],_Inst).
instantiated([Var|Rest],Inst) :-
	memberchk(Var-_Val,Inst),
	instantiated(Rest,Inst).

/*---------------------------------------------------------------------
  Predicate: var_only_unist(+BN,+Var,+Inst,+InVarFacts,+InRest,-OutVarFacts,-OutRest)
       Args: 
   Succeeds: exactly once
   Comments: Pull out all those factors which only need Var to be instantiated
             to become fully instantiated
---------------------------------------------------------------------*/
var_only_unist([],_Var,_Inst,VarFacts,Rest,VarFacts,Rest).
var_only_unist([Child-Parents|T],Var,Inst,InVarFacts,InRest,OutVarFacts,OutRest) :-
	select(Var,[Child|Parents],Others),  
	instantiated(Others,Inst),
	!,
	var_only_unist(T,Var,Inst,[Child-Parents|InVarFacts],InRest,OutVarFacts,OutRest).
var_only_unist([Family|T],Var,Inst,InVarFacts,InRest,OutVarFacts,OutRest) :-
	var_only_unist(T,Var,Inst,InVarFacts,[Family|InRest],OutVarFacts,OutRest).

/*---------------------------------------------------------------------
  Predicate: map_memberchk(+Vars,+Inst,-Vals)
       Args: Vals: the values Inst gives Vars
             Vars: variable names
	     Inst: An instantiation of variables in BN of the form
	     [var3-val13,var4-val42..]
   Succeeds: at most once (but should always succeed)
   Comments: 
---------------------------------------------------------------------*/
map_memberchk([],_Inst,[]).
map_memberchk([Parent|Rest],Inst,[Val|Vals]) :-
	memberchk(Parent-Val,Inst),
	map_memberchk(Rest,Inst,Vals).










/*----------------------------------------------------------------------

example BN, and instantiation

----------------------------------------------------------------------*/

%order of parents must match cpt/3
bn([s1-[m],s2-[m],s3-[m,f],m-[],f-[]]).

inst([s1-[1],s2-[1,2],s3-[1,2],m-[2],f-[2]]).

/*---------------------------------------------------------------------
  Predicate: cpt(+ChildInst,+ParentsInst,-Prob)
       Args: 
   Succeeds: exactly once
   Comments: All instantiations are assumed complete
---------------------------------------------------------------------*/

%example cpt function

%cpt(s1-1,[]-[],0.3) :- !.
%cpt(s1-2,[]-[],0.7) :- !.
%cpt(s1-1,[m]-[2],0.3) :- !.
%cpt(s1-2,[m]-[2],0.7) :- !.
%cpt(_,_,0.5).			% py = 0.5 for all else

%order of values must match instantiation
%all_vals(_Var,[1,2]).
