:- module( degrees, [
	degree_of_inbreeding/2,
	degree_of_promiscuity/2,
	degree_of_longest_chain/2,
	flatten_1/2,
	kv_decompose_l/3
	] ).

:- ensure_loaded( library(lists) ).		%.
:- ensure_loaded( library(ugraphs) ).		%.
:- ensure_loaded( library(ordsets) ).		%.

/*
compile( degrees), save_modules( [degrees], degrees ), halt.
*/

degree_of_inbreeding( Ped, I ) :-
	degree_of_inbreeding( Ped, Ped, 0, I ).

degree_of_inbreeding( [], _Ped, I, I ).
degree_of_inbreeding( [_Idv-[Pa1,Pa2]|T], Ped, AccI, I ) :-
	% individual_is_inbred( Pa1, Pa2, Ped ),
	reachable( Pa1, Ped, Anc1 ),
	% reachable( Pa1, Ped, PrvAnc1 ),
	% ord_del_element( PrvAnc1, Pa1, Anc1 ),
	reachable( Pa2, Ped, Anc2 ),
	% reachable( Pa2, Ped, PrvAnc2 ),
	% ord_del_element( PrvAnc2, Pa2, Anc2 ),
	ord_intersect( Anc1, Anc2 ),
	% write( inbreed(Idv,Pa1,Pa2) ), nl,
	!,
	NxI is AccI + 1,
	degree_of_inbreeding( T, Ped, NxI, I ).
degree_of_inbreeding( [_|T], Ped, AccI, I ) :-
	degree_of_inbreeding( T, Ped, AccI, I ).

individual_is_inbred( _Idv, _Pa, _Pa1 ).
individual_is_inbred( [HIdv-HPa|T], Ped, Idv, Pa1, AccI, I ) :-
	increases_inbredding_counter( HIdv, HPa, Idv, Ped, AccI, NxAccI ),
	individual_degree_of_inbreeding( T, Ped, Idv, Pa1, NxAccI, I ).

increases_inbredding_counter( HIdv, HPa, Idv, Ped, AccI, NxAccI ) :-
	ord_member( HIdv, Pa1 ),
	ord_del_element(HPa,Idv,NoIPa),
	parents_have_common_ancestor( NoIPa, Pa1, Ped ),
	!,
	NxAccI is AccI + 1.
increases_inbredding_counter( _HIdv, _Idv, _Ped, AccI, AccI ).

parents_have_common_ancestor( Pa1, Pa2, Ped ) :-
	member( I1, Pa1 ),
	member( I2, Pa2 ),
	reachable( I1, Ped, PrvAnc1 ),
	ord_del_element( PrvAnc1, I1, Anc1 ),
	reachable( I2, Ped, PrvAnc2 ),
	ord_del_element( PrvAnc2, I2, Anc2 ),
	ord_intersect( Anc1, Anc2 ),
	!.

degree_of_promiscuity( Ped, P ) :-
	degree_of_promiscuity( Ped, 0, P ).

degree_of_promiscuity( [], P, P ).
degree_of_promiscuity( [_Idv-Pa|T], AccP, P ) :-
	individual_degree_of_promiscuity( T, Pa, AccP, NxP ),
	degree_of_promiscuity( T, NxP, P ).

individual_degree_of_promiscuity( [], _Pa1, P, P ).
individual_degree_of_promiscuity( [_Idv-HPa|T], Pa1, AccP, P ) :-
	ord_intersection( HPa, Pa1, ComPa ),
	increases_degree_of_promiscuity( ComPa, AccP, NxP ),
	individual_degree_of_promiscuity( T, Pa1, NxP, P ).

increases_degree_of_promiscuity( [_], AccP, NxP ) :-
	!,
	NxP is AccP + 1.
increases_degree_of_promiscuity( [_], AccP, AccP ).

degree_of_longest_chain( Ped, G ) :-
	top_sort( Ped, TSorted ),
	reverse( TSorted, Reversed ),
	degree_of_longest_chain( Reversed, Ped, [], 0, G ).

degree_of_longest_chain( [], _EmptyPed, _, G, G ).
degree_of_longest_chain( [Idv|T], Ped, Found, AccG, G ) :-
	select( Idv-Pa, Ped, RedPed ),
	individual_degree_of_longest_chain( Pa, Found, 0, IdvG ),
	ord_add_element( Found, Idv-IdvG, NxFound ),
	% ( IdvG > AccG -> NxG is IdvG ; NxG is AccG ),
	NxG is max( IdvG, AccG ),
	degree_of_longest_chain( T, RedPed, NxFound, NxG, G ).

individual_degree_of_longest_chain( [], _Found, G, G ).
individual_degree_of_longest_chain( [H|T], Found, AccG, G ) :-
	memberchk( H-HG, Found ),
	% ( memberchk(H-HG,Found) -> true  ; 
		% write( user_error, not_in_found(H,Found) ), 
		% nl( user_error ), abort ),
	HGinc is HG + 1,
	NxG is max( HGinc, AccG ),
	% ( AccG < HG + 1 -> NxG is HG + 1 ; NxG is AccG ),
	individual_degree_of_longest_chain( T, Found, NxG, G ).

flatten_1( Nlist, Flist ) :-
	flatten_1( Nlist, Flist, [] ).
flatten_1( [], L0, L ) :-
	!,
	L0 = L.
flatten_1( [H|T], L0, L ) :-
	!,
	flatten_1( H, L0, L1 ),
	flatten_1( T, L1, L ).
flatten_1( Oth, [Oth|List], List ).

kv_decompose_l( [], [], [] ).
kv_decompose_l( [K-V|T], [K|KT], [V|VT] ) :-
	kv_decompose_l( T, KT, VT ).
